#include <eigen3/Eigen/Dense>
#include <eigen3/Eigen/Sparse>
#include <eigen3/Eigen/SparseLU>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <functional>
#include <vector>
#include <chrono>
using namespace std::chrono;
using namespace Eigen;

//! Solving the possibly overdetermined linear system of equations
//! Ax = b by solving normal equations
//! The least squares solution is returned
////TASK A
VectorXd normalEquationSolve(const MatrixXd &A, const VectorXd &b){
    VectorXd x;
    // TO-DO task (a)
    // ..
    // ..
    // ..
    return x ; //return solution vector
}
////END OF TASK A

////TASK C
////TASK C_1
//! Given a sparse matrix A, construct the sparse matrix 'C' for the modified system Cy=d
SparseMatrix<double> computeC(SparseMatrix<double> &A, const VectorXd &b){
    SparseMatrix <double> C;
    // TO-DO task (c)
    // ..
    // ..
    // ..
    return C;
}
////END OF TASK C_1

////TASK C_2
//! Given the original RHS-vector b, construct the RHS-vector 'd' for the modified system Cy=d
VectorXd computeRHS(SparseMatrix<double> &A, const VectorXd &b){
    VectorXd d;
    // TO-DO task (c)
    // ..
    // ..
    // ..
    return d ;
}
////END OF TASK C_2
////END OF TASK C

////TASK D
//Obtain solution for the modified system using a SparseLU solver
VectorXd computeSolution(SparseMatrix<double> &A, const VectorXd &b){
    VectorXd x;
    // TO-DO task (d)
    // ..
    // ..
    // ..
    return x; //return the actual solution vector 'x' for the original LSE Ax=b
}
////END OF TASK D

double f(double x) {
    return M_PI*M_PI*std::sin(M_PI*x); //returns value of RHS-function f at x
}

////TASK F
////TASK F_1
//Construct dense matrix 'A' obtained through Finite Difference Method for the poisson problem
MatrixXd constructDenseA_Poisson(int N){
    MatrixXd A_dense;
    // TO-DO task (f)
    // ..
    // ..
    // ..
    return A_dense;
}
////END OF TASK F_1

////TASK F_2
//Construct sparse matrix 'A' obtained through Finite Difference Method for the poisson problem
SparseMatrix <double> constructSparseA_Poisson(MatrixXd &A_dense, int N){
	SparseMatrix <double> A_sparse;
	// TO-DO task (f)
	// ..
	return A_sparse;
}
////END OF TASK F_2
////END OF TASK F

int main() {
    // Mesh size (number of interior points)
    std::vector<int> N_interior = {25, 50, 75, 100, 200, 400, 800, 1600};
    for(unsigned int p = 0; p < N_interior.size(); ++p) {
        int N  = N_interior.at(p);
        double delta_x = 1.0/(N+1);

        VectorXd x(N+2), b(N), solution_exact(N);
        
        for(unsigned int i = 0; i < N + 2; ++i) {
            x(i) = i * delta_x; //Construct mesh vector
            
            if (i > 0 && i < N+1) {
                solution_exact(i-1) = std::sin(M_PI*x(i)); //Construct solution vector
                b(i-1) = f(x(i)); //Construct RHS vector 'b'
            }
        }
        
        
    //Obtain solution by standard normal equation and measure run-time
        MatrixXd A_dense = constructDenseA_Poisson(N);
        auto tic_normal = high_resolution_clock::now();
        VectorXd solution_normal = normalEquationSolve(A_dense, b);
        auto toc_normal = high_resolution_clock::now();
        duration<double> runTime_normal = duration_cast<duration<double>>(toc_normal-tic_normal);
        
    //Obtain solution by modified system which preserves sparsity and measure run-time
        SparseMatrix <double> A_sparse = constructSparseA_Poisson(A_dense, N);
        auto tic_sparse = high_resolution_clock::now();
        VectorXd solution_sparse = computeSolution(A_sparse, b);
        auto toc_sparse = high_resolution_clock::now();
        duration<double> runTime_sparse = duration_cast<duration<double>>(toc_sparse-tic_sparse);
        
    // Compute error in solution for both methods
        double maxerr_normal = (solution_exact - solution_normal).template lpNorm<Infinity>();
        double maxerr_sparse = (solution_exact - solution_sparse).template lpNorm<Infinity>();
        
    // I/O
        if (p == 0) std::cout << std::setw(15) << "N" << std::setw(15) << "maxerr_normal" << std::setw(15) << "runTime_normal" << std::setw(15) << "maxerr_sparse" << std::setw(15) << "runTime_sparse" << std::endl;
        std::cout << std::setw(15) << N << std::setw(15) << maxerr_normal << std::setw(15) << runTime_normal.count() << std::setw(15) << maxerr_sparse << std::setw(15) << runTime_sparse.count() << std::endl;
    }
}


// End of file
